﻿using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace TestProject;

/// <summary>
/// 확대/축소 컨트롤
/// </summary>
public class ZoomableCanvas : Control
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Public

    #region 캔버스 너비 속성 - CanvasWidthProperty

    /// <summary>
    /// 캔버스 너비 속성
    /// </summary>
    public static readonly DependencyProperty CanvasWidthProperty = DependencyProperty.Register
    (
        "CanvasWidth",
        typeof(double),
        typeof(ZoomableCanvas),
        new PropertyMetadata(300.0, CanvasSizePropertyChangedCallback)
    );

    #endregion
    #region 캔버스 높이 속성 - CanvasHeightProperty

    /// <summary>
    /// 캔버스 높이 속성
    /// </summary>
    public static readonly DependencyProperty CanvasHeightProperty = DependencyProperty.Register
    (
        "CanvasHeight",
        typeof(double),
        typeof(ZoomableCanvas),
        new PropertyMetadata(200.0, CanvasSizePropertyChangedCallback)
    );

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Property
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 캔버스 너비 - CanvasWidth

    /// <summary>
    /// 캔버스 너비
    /// </summary>
    public double CanvasWidth
    {
        get => (double)GetValue(CanvasWidthProperty);
        set => SetValue(CanvasWidthProperty, value);
    }

    #endregion
    #region 캔버스 높이 - CanvasHeight

    /// <summary>
    /// 캔버스 높이
    /// </summary>
    public double CanvasHeight
    {
        get => (double)GetValue(CanvasHeightProperty);
        set => SetValue(CanvasHeightProperty, value);
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Field
    ////////////////////////////////////////////////////////////////////////////////////////// Private

    #region Field

    /// <summary>
    /// 캔버스
    /// </summary>
    private Canvas canvas;

    /// <summary>
    /// 스케일 변환
    /// </summary>
    private ScaleTransform scaleTransform;

    /// <summary>
    /// 이동 변환
    /// </summary>
    private TranslateTransform translateTransform;

    /// <summary>
    /// 시작 포인트
    /// </summary>
    private Point startPoint;

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
    ////////////////////////////////////////////////////////////////////////////////////////// Static

    #region 생성자 - ZoomableCanvas()

    /// <summary>
    /// 생성자
    /// </summary>
    static ZoomableCanvas()
    {
        DefaultStyleKeyProperty.OverrideMetadata(typeof(ZoomableCanvas), new FrameworkPropertyMetadata(typeof(ZoomableCanvas)));
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Private

    #region 캔버스 크기 속성 변경시 콜백 처리하기 - CanvasSizePropertyChangedCallback(d, e)

    /// <summary>
    /// 캔버스 크기 속성 변경시 콜백 처리하기
    /// </summary>
    /// <param name="d">이벤트 발생자</param>
    /// <param name="e">이벤트 인자</param>
    private static void CanvasSizePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
    {
        ZoomableCanvas zoomableCanvas = d as ZoomableCanvas;

        if(zoomableCanvas != null && zoomableCanvas.canvas != null)
        {
            zoomableCanvas.canvas.Width  = zoomableCanvas.CanvasWidth;
            zoomableCanvas.canvas.Height = zoomableCanvas.CanvasHeight;
        }
    }

    #endregion

    ////////////////////////////////////////////////////////////////////////////////////////// Instance
    //////////////////////////////////////////////////////////////////////////////// Public

    #region 템플리트 적용시 처리하기 - OnApplyTemplate()

    /// <summary>
    /// 템플리트 적용시 처리하기
    /// </summary>
    public override void OnApplyTemplate()
    {
        base.OnApplyTemplate();

        this.canvas = GetTemplateChild("PART_Canvas") as Canvas;

        if(this.canvas != null)
        {
            this.scaleTransform     = new ScaleTransform();
            this.translateTransform = new TranslateTransform();

            TransformGroup transformGroup = new TransformGroup();

            transformGroup.Children.Add(this.scaleTransform    );
            transformGroup.Children.Add(this.translateTransform);

            this.canvas.RenderTransformOrigin = new Point(0.5, 0.5);
            this.canvas.RenderTransform       = transformGroup;
        }
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////// Protected

    #region 마우스 휠 처리하기 - OnMouseWheel(e)

    /// <summary>
    /// 마우스 휠 처리하기
    /// </summary>
    /// <param name="e">이벤트 인자</param>
    protected override void OnMouseWheel(MouseWheelEventArgs e)
    {
        base.OnMouseWheel(e);

        if(this.canvas == null)
        {
            return;
        }

        double zoom = e.Delta > 0 ? 1.1 : 0.9;

        Point mousePoition = e.GetPosition(this.canvas);

        this.scaleTransform.ScaleX *= zoom;
        this.scaleTransform.ScaleY *= zoom;

        this.translateTransform.X += (1 - zoom) * mousePoition.X;
        this.translateTransform.Y += (1 - zoom) * mousePoition.Y;
    }

    #endregion
    #region 마우스 왼쪽 버튼 DOWN 처리하기 - OnMouseLeftButtonDown(e)

    /// <summary>
    /// 마우스 왼쪽 버튼 DOWN 처리하기
    /// </summary>
    /// <param name="e">이벤트 인자</param>
    protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
    {
        base.OnMouseLeftButtonDown(e);

        this.startPoint = e.GetPosition(this);

        CaptureMouse();
    }

    #endregion
    #region 마우스 이동시 처리하기 - OnMouseMove(e)

    /// <summary>
    /// 마우스 이동시 처리하기
    /// </summary>
    /// <param name="e">이벤트 인자</param>
    protected override void OnMouseMove(MouseEventArgs e)
    {
        base.OnMouseMove(e);

        if(IsMouseCaptured)
        {
            Point currentPoint = e.GetPosition(this);

            Vector deltaVector = currentPoint - this.startPoint;

            this.translateTransform.X += deltaVector.X;
            this.translateTransform.Y += deltaVector.Y;

            this.startPoint = currentPoint;
        }
    }

    #endregion
    #region 마우스 왼쪽 버튼 UP 처리하기 - OnMouseLeftButtonUp(e)

    /// <summary>
    /// 마우스 왼쪽 버튼 UP 처리하기
    /// </summary>
    /// <param name="e">이벤트 인자</param>
    protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
    {
        base.OnMouseLeftButtonUp(e);

        ReleaseMouseCapture();
    }

    #endregion
}